/**
 * This boot file initalizes the store from persistent storage and load all plugins
 */
import {
  PersistentStorage,
  api,
  isAxiosError,
  saveSession,
  useMainStore,
} from '@flaschengeist/api';
import { Notify, Platform } from 'quasar';
import { loadPlugins } from './plugins';
import { boot } from 'quasar/wrappers';
import routes from 'src/router/routes';

async function loadBaseUrl() {
  try {
    const url = await PersistentStorage.get<string>('baseURL');
    if (url !== null) api.defaults.baseURL = url;
  } catch (e) {
    console.warn('Could not load BaseURL', e);
  }
}

class BackendError extends Error {}

/**
 * Loading backend information
 * @returns Backend object or null
 */
async function getBackend() {
  const { data } = await api.get<FG.Backend>('/');
  if (!data || typeof data !== 'object' || !('plugins' in data))
    throw new BackendError('Invalid backend response received');
  return data;
}

/**
 * Boot file for loading baseURL + Session from PersistentStorage + loading and initializing all plugins
 */
export default boot(async ({ app, router }) => {
  const store = useMainStore();

  // FIRST(!) get the base URL
  await loadBaseUrl();

  // Init the store, load current session and user, if available
  try {
    await store.init();
  } finally {
    // Any changes on the session is written back to the persistent store
    store.$subscribe((mutation, state) => {
      saveSession(state.session);
    });
  }

  // Load all plugins
  try {
    // Fetch backend data
    const backend = await getBackend();
    // Load enabled plugins
    const flaschengeist = await loadPlugins(backend, routes);
    // Add loaded routes to router
    flaschengeist.routes.forEach((route) => router.addRoute(route));
    // save plugins in VM-variable
    app.provide('flaschengeist', flaschengeist);
  } catch (error) {
    // Handle errors from loading the backend information
    if (error instanceof BackendError || isAxiosError(error)) {
      router.isReady().finally(() => {
        if (Platform.is.capacitor) void router.push({ name: 'setup_backend' });
        else void router.push({ name: 'offline', params: { refresh: 1 } });
      });
    } else if (typeof error === 'string') {
      // Handle plugin not found errors
      void router.push({ name: 'error' });
      Notify.create({
        type: 'negative',
        message: `Fehler beim Laden: Bitte wende dich an den Admin (${error})!`,
        timeout: 10000,
        progress: true,
      });
    } else {
      console.error('Unknown error in init.ts:', error);
    }
  }
});
